@@ -4,14 +4,13 @@
  * Parts of this file are derived from apt.  Apt is copyright 1997,
  * 1998, 1999 Jason Gunthorpe and others.
  *
- * Copyright (C) 2005 Nokia Corporation.
+ * Copyright (C) 2005, 2006 Nokia Corporation.  All Right reserved.
  *
  * Contact: Marius Vollmer <marius.vollmer@nokia.com>
  *
  * This program is free software; you can redistribute it and/or
- * modify it under the terms of the GNU General Public License as
- * published by the Free Software Foundation; either version 2 of the
- * License, or (at your option) any later version.
+ * modify it under the terms of the GNU General Public License version
+ * 2 as published by the Free Software Foundation.
  *
  * This program is distributed in the hope that it will be useful, but
  * WITHOUT ANY WARRANTY; without even the implied warranty of
@@ -79,6 +78,7 @@
 #include <apt-pkg/debversion.h>
 #include <apt-pkg/dpkgpm.h>
 #include <apt-pkg/debsystem.h>
+#include <apt-pkg/orderlist.h>
 
 #include <glib/glist.h>
 #include <glib/gstring.h>
@@ -99,8 +99,9 @@ using namespace std;
 
    COMMUNICATING WITH THE FRONTEND
 
-   COMMAND DISPATCHER
+   STARTUP AND COMMAND DISPATCHER
 
+   CACHE HANDLING
 */
 
 
@@ -191,6 +192,7 @@ must_open (char *filename, int flags)
       perror (filename);
       exit (1);
     }
+  return fd;
 }
 
 /* MAYBE_READ_BYTE reads a byte from FD if one is available without
@@ -229,8 +231,8 @@ drain_fd (int fd)
     ;
 }
 
-/* Get a lock as with GetLock, breaking it if needed and allowed
-   by flag_break_locks.
+/* Get a lock as with GetLock from libapt-pkg, breaking it if needed
+   and allowed by flag_break_locks.
 
    We do this so that the users can not lock themselves out.  We break
    locks instead of not locking since noisily breaking a lock is
@@ -259,9 +261,9 @@ ForceLock (string File, bool Errors = true)
 /** COMMUNICATING WITH THE FRONTEND.
  
    The communication with the frontend happens over four
-   unidirectional fifos: requests are read from INPUT_FD and
-   responses are sent back via OUTPUT_FD.  No new requests is read
-   until the response to the current one has been completely sent.
+   unidirectional fifos: requests are read from INPUT_FD and responses
+   are sent back via OUTPUT_FD.  No new request is read until the
+   response to the current one has been completely sent.
 
    The data read from INPUT_FD must follow the request format
    specified in <apt-worker-proto.h>.  The data written to OUTPUT_FD
@@ -272,12 +274,12 @@ ForceLock (string File, bool Errors = true)
    currently no meaning defined for the actual bytes that are sent,
    the mere arrival of a byte triggers the abort.
 
-   When using the apt-pkg PackageManager, it is configured in such a
-   way that it sents it "pmstatus:" message lines to STATUS_FD.
+   When using the libapt-pkg PackageManager, it is configured in such
+   a way that it sends it "pmstatus:" message lines to STATUS_FD.
    Other asynchronous status reports are sent as spontaneous
    APTCMD_STATUS responses via OUTPUT_FD.  'Spontaneous' should mean
-   that no request is required to receive APTCMD_STATUS responses.
-   In fact, APTCMD_STATUS requests are treated as an error by the
+   that no request is required to receive APTCMD_STATUS responses.  In
+   fact, APTCMD_STATUS requests are treated as an error by the
    apt-worker.
 
    Logging and debug output, and output from dpkg and the maintainer
@@ -322,7 +324,7 @@ must_read (void *buf, size_t n)
 }
 
 static void
-must_write (void *buf, size_t n)
+must_write (void *buf, ssize_t n)
 {
   if (n > 0 && write (output_fd, buf, n) != n)
     {
@@ -409,9 +411,20 @@ void cmd_clean ();
 void cmd_get_file_details ();
 void cmd_install_file ();
 
+/* Commands can request the package cache to be refreshed by calling
+   NEED_CACHE_INIT before they return.  The cache will then be
+   reconstructed after sending the response and before starting to
+   handle the next command.  In this way, the cache reconstruction
+   happens in the background.
+
+   XXX - However, APTCMD_STATUS messages are still being sent when the
+         cache is reconstructed in the background and the UI has some
+         ugly logic to deal with that.
+*/
+
 static bool init_cache_after_request;
 
-void cache_init ();
+void cache_init (bool with_status = true);
 
 void
 need_cache_init ()
@@ -511,7 +524,7 @@ handle_request ()
   free_buf (reqbuf, stack_reqbuf);
 
   if (init_cache_after_request)
-    cache_init ();
+    cache_init (false);
 }
 
 void read_certified_conf ();
@@ -550,14 +563,58 @@ main (int argc, char **argv)
   if (nice (20) == -1 && errno != 0)
     log_stderr ("nice: %m");
 
-  cache_init ();
+  cache_init (false);
   read_certified_conf ();
 
   while (true)
     handle_request ();
 }
 
-/** COMMAND HANDLERS
+/** CACHE HANDLING
+
+    This section contains some general purpose functions to maintain
+    the cache of the package database.
+
+    The package cache can represent both the 'current' situation
+    (i.e., the union of the information from /var/lib/dpkg/status and
+    the various Packages files downloaded from repositories) and a
+    'desired' situation.
+
+    A operation such as installing a package is performed by modifying
+    the 'desired' situation in the cache and if that leads to a
+    consistent configuration, the 'current' situation is brought in
+    line with the 'desired' one by downloading the needed archives and
+    running dpkg in an approriate way.
+
+    We have our own idea of what should happen when a new package (or
+    a new version of a package) is installed, for example, and the
+    functions in this section implement this idea.  These principal
+    functions are available:
+
+    - cache_init
+
+    This function creates or recreates the cache from
+    /var/lib/dpkg/status and the various Packages file from the
+    repositories.
+
+    - cache_reset ()
+
+    This function resets the 'desired' state of the cache to be
+    identical to the 'current' one.
+
+    - mark_for_install ()
+
+    This function modifies the 'desired' state of the cache to reflect
+    the installation of the given package.  It will try to achieve a
+    consistent 'desired' configuration by installing mising
+    dependencies etc.  In general, it implements our own installation
+    smartness.
+
+    - mark_for_remove ()
+
+    This function modifies the 'desired' state of the cache to reflect
+    the removal of the given package.  As with mark_for_install,
+    mark_for_removal implements our own removal smartness.
  */
 
 /* We only report real progress information when reconstructing the
@@ -571,11 +628,17 @@ main (int argc, char **argv)
 
 class UpdateProgress : public OpProgress
 {
+  bool with_status;
+
   virtual void
   Update ()
   {
-    send_status (op_updating_cache, (int)Percent, 100, 5);
+    if (with_status)
+      send_status (op_updating_cache, (int)Percent, 100, 5);
   }
+
+public:
+  UpdateProgress (bool ws) : with_status (ws) { }
 };
 
 class DownloadStatus : public pkgAcquireStatus
@@ -636,11 +699,16 @@ is_user_package (const pkgCache::VerIterator &ver)
          pkgCache, a pkgDebCache, etc.
 */
 pkgCacheFile *package_cache = NULL;
-char *auto_flags = NULL;
-unsigned int package_count;
 
-/* Save and restore the 'Auto' flags of the cache.  Libapt-pkg does
-   not seem to save it so we do it.
+unsigned int package_count;
+struct package_flag_struct {
+  bool autoinst, related;
+} *package_flags = NULL;
+
+/* Save the 'Auto' flags of the cache.  Libapt-pkg does not seem to
+   save it so we do it.  We also make a copy of the Auto flags in our
+   own PACKAGE_FLAGS storage so that CACHE_RESET will reset the Auto
+   flags to the state last saved with this function.
 */
 
 void
@@ -657,39 +725,29 @@ save_auto_flags ()
       return;
     }
 
-  memset (auto_flags, 0, package_count);
-
   FILE *f = fopen ("/var/lib/osso-application-installer/autoinst", "w");
   if (f)
     {
       pkgDepCache &cache = *package_cache;
 
       for (pkgCache::PkgIterator pkg = cache.PkgBegin(); !pkg.end (); pkg++)
-	if (cache[pkg].Flags & pkgCache::Flag::Auto)
-	  {
-	    auto_flags[pkg->ID] = true;
-	    fprintf (f, "%s\n", pkg.Name ());
-	  }
+	{
+	  if (cache[pkg].Flags & pkgCache::Flag::Auto)
+	    {
+	      package_flags[pkg->ID].autoinst = true;
+	      fprintf (f, "%s\n", pkg.Name ());
+	    }
+	  else
+	    package_flags[pkg->ID].autoinst = false;
+	}
       fclose (f);
     }
 }
 
-void
-restore_auto_flags ()
-{
-  if (package_cache == NULL)
-    return;
-
-  pkgDepCache &cache = *package_cache;
-  
-  for (pkgCache::PkgIterator pkg = cache.PkgBegin(); !pkg.end (); pkg++)
-    {
-      if (auto_flags[pkg->ID])
-	cache[pkg].Flags |= pkgCache::Flag::Auto;
-      else
-	cache[pkg].Flags &= ~pkgCache::Flag::Auto;
-    }
-}
+/* Load the Auto flags and put them into our own PACKAGE_FLAGS
+   storage.  You need to call CACHE_RESET to transfer them into the
+   actual cache.
+*/
 
 void
 load_auto_flags ()
@@ -699,7 +757,8 @@ load_auto_flags ()
   if (package_cache == NULL)
     return;
 
-  memset (auto_flags, 0, package_count);
+  for (unsigned int i = 0; i < package_count; i++)
+    package_flags[i].autoinst = false;
 
   FILE *f = fopen ("/var/lib/osso-application-installer/autoinst", "r");
   if (f)
@@ -719,19 +778,18 @@ load_auto_flags ()
 	  if (!pkg.end ())
 	    {
 	      DBG ("auto: %s", pkg.Name ());
-	      auto_flags[pkg->ID] = true;
+	      package_flags[pkg->ID].autoinst = true;
 	    }
 	}
 
       free (line);
       fclose (f);
-
-      restore_auto_flags ();
     }
 }
 
 /* Our own version of debSystem.  We override the Lock member function
-   to be able to break locks and clear the updates journal of dpkg.
+   to be able to break locks and to avoid failing when dpkg has left a
+   journal.
 */
 
 class mydebSystem : public debSystem
@@ -739,7 +797,6 @@ class mydebSystem : public debSystem
   // For locking support
   int LockFD;
   unsigned LockCount;
-  void ClearUpdates ();
 
 public:
 
@@ -762,43 +819,6 @@ mydebSystem::Score (Configuration const &Cnf)
   return debSystem::Score (Cnf) + 10;  // Pick me, pick me, pick me!
 }
 
-void
-mydebSystem::ClearUpdates()
-{
-  string File = flNotFile(_config->Find("Dir::State::status")) + "updates/";
-  DIR *DirP = opendir(File.c_str());
-  if (DirP == 0)
-    return;
-   
-  /* We ignore any files that are not all digits, this skips .,.. and 
-     some tmp files dpkg will leave behind.. */
-
-  for (struct dirent *Ent = readdir(DirP); Ent != 0; Ent = readdir(DirP))
-    {
-      bool ignore = false;
-
-      for (unsigned int I = 0; Ent->d_name[I] != 0; I++)
-	{
-	  // Check if its not a digit..
-	  if (isdigit(Ent->d_name[I]) == 0)
-	    {
-	      ignore = true;
-	      break;
-	    }
-	}
-
-      if (!ignore)
-	{
-	  string name = File + Ent->d_name;
-	  log_stderr ("Cleaning %s", name.c_str ());
-	  if (unlink (name.c_str()) < 0 && errno != ENOENT)
-	    log_stderr ("%s: %m", name.c_str ());
-	}
-    }
-
-   closedir(DirP);
-}
-
 bool
 mydebSystem::Lock ()
 {
@@ -806,8 +826,6 @@ mydebSystem::Lock ()
      copy of the behavior of dpkg.
   */
 
-  bool tried_to_rescue_already = false;
-
   // Disable file locking
   if (_config->FindB("Debug::NoLocking",false) == true || LockCount > 1)
     {
@@ -828,11 +846,6 @@ mydebSystem::Lock ()
 			     "are you root?",AdminDir.c_str());
     }
 
-  /* Clear out the dpkg journal.
-   */
-
-  ClearUpdates();
-
   LockCount++;
       
   return true;
@@ -859,12 +872,51 @@ bool mydebSystem::UnLock(bool NoErrors)
 									/*}}}*/
 mydebSystem mydebsystem;
 
+static void
+clear_dpkg_updates ()
+{
+  string File = flNotFile(_config->Find("Dir::State::status")) + "updates/";
+  DIR *DirP = opendir(File.c_str());
+  if (DirP == 0)
+    return;
+   
+  /* We ignore any files that are not all digits, this skips .,.. and 
+     some tmp files dpkg will leave behind.. */
+
+  for (struct dirent *Ent = readdir(DirP); Ent != 0; Ent = readdir(DirP))
+    {
+      bool ignore = false;
+
+      for (unsigned int I = 0; Ent->d_name[I] != 0; I++)
+	{
+	  // Check if its not a digit..
+	  if (isdigit(Ent->d_name[I]) == 0)
+	    {
+	      ignore = true;
+	      break;
+	    }
+	}
+
+      if (!ignore)
+	{
+	  log_stderr ("Running 'dpkg --configure dpkg' "
+		      "to clean up the journal.");
+	  system ("dpkg --configure dpkg");
+	  break;
+	}
+    }
+
+   closedir(DirP);
+}
+
+void cache_reset ();
+
 /* Initialize libapt-pkg if this has not been already and (re-)create
    PACKAGE_CACHE.  If the cache can not be created, PACKAGE_CACHE is
    set to NULL and an appropriate message is output.
 */
 void
-cache_init ()
+cache_init (bool with_status)
 {
   static bool global_initialized = false;
 
@@ -873,7 +925,9 @@ cache_init ()
       DBG ("closing");
       package_cache->Close ();
       delete package_cache;
+      delete[] package_flags;
       package_cache = NULL;
+      package_flags = NULL;
       DBG ("done");
     }
 
@@ -892,8 +946,12 @@ cache_init ()
      cause the following operations to fail.
   */
   _error->DumpErrors ();
+
+  /* Clear out the dpkg journal before construction the cache.
+   */
+  clear_dpkg_updates ();
 	  
-  UpdateProgress progress;
+  UpdateProgress progress (with_status);
   package_cache = new pkgCacheFile;
 
   DBG ("init.");
@@ -905,21 +963,15 @@ cache_init ()
       package_cache = NULL;
     }
 
-  if (auto_flags)
-    {
-      delete[] auto_flags;
-      auto_flags = NULL;
-    }
-
   if (package_cache)
     {
       pkgDepCache &cache = *package_cache;
       package_count = cache.Head().PackageCount;
-      auto_flags = new char[package_count];
-      memset (auto_flags, 0, package_count);
+      package_flags = new package_flag_struct[package_count];
     }
 
   load_auto_flags ();
+  cache_reset ();
 }
 
 /* Determine whether a package was installed automatically to satisfy
@@ -933,11 +985,49 @@ is_auto_package (pkgCache::PkgIterator &pkg)
   return (cache[pkg].Flags & pkgCache::Flag::Auto) != 0;
 }
 
-/* Revert the cache to its initial state.  More concretely, all
-   packages are marked as 'keep'.
+/* Determine whether a package is related to the current operation.
+*/
+bool
+is_related (pkgCache::PkgIterator &pkg)
+{
+  return package_flags[pkg->ID].related;
+}
 
-   The effect should be the same as calling pkgDebCache::Init, but
-   much faster.
+void
+mark_related (pkgCache::PkgIterator &pkg)
+{
+  if (package_flags[pkg->ID].related)
+    return;
+
+  package_flags[pkg->ID].related = true;
+
+  pkgDepCache &cache = *package_cache;
+
+  if (pkg.State() == pkgCache::PkgIterator::NeedsUnpack)
+    cache.SetReInstall (pkg, true);
+  
+  pkgCache::VerIterator cand = cache.GetCandidateVer (pkg);
+  if (cand.end ())
+    return;
+
+  for (pkgCache::DepIterator dep = cand.DependsList(); dep.end() == false;
+       dep++)
+    {
+      if (dep->Type == pkgCache::Dep::PreDepends ||
+	  dep->Type == pkgCache::Dep::Depends)
+	{
+	  pkgCache::PkgIterator p = dep.TargetPkg ();
+	  if (!p.end ())
+	    mark_related (p);
+	}
+    }
+}
+
+/* Revert the cache to its initial state.  More concretely, all
+   packages are marked as 'keep' and 'unrelated'.
+
+   The effect on the cache should be the same as calling
+   pkgDebCache::Init, but much faster.
 */
 void
 cache_reset_package (pkgCache::PkgIterator &pkg)
@@ -945,10 +1035,12 @@ cache_reset_package (pkgCache::PkgIterator &pkg)
   pkgDepCache &cache = *package_cache;
 
   cache.MarkKeep (pkg);
-  if (auto_flags[pkg->ID])
+  if (package_flags[pkg->ID].autoinst)
     cache[pkg].Flags |= pkgCache::Flag::Auto;
   else
     cache[pkg].Flags &= ~pkgCache::Flag::Auto;
+
+  package_flags[pkg->ID].related = false;
 }
 
 void
@@ -982,8 +1074,41 @@ mark_for_install (pkgCache::PkgIterator &pkg)
       if (cache[p].Delete ())
 	cache_reset_package (p);
     }
+
+  mark_related (pkg);
 }
 
+/* Mark every upgradeable non-user package for installation.
+ */
+static void
+mark_sys_upgrades ()
+{
+  pkgDepCache &cache = *package_cache;
+
+  for (pkgCache::PkgIterator p = cache.PkgBegin (); !p.end (); p++)
+    {
+      if (!p.CurrentVer().end() && !is_user_package (p.CurrentVer()))
+	mark_for_install (p);
+    }
+}
+
+/* Mark the named package for installation.  This function also
+   handles magic packages like "magic:sys".
+*/
+
+static void
+mark_named_package_for_install (const char *package)
+{
+  if (!strcmp (package, "magic:sys"))
+    mark_sys_upgrades ();
+  else
+    {
+      pkgDepCache &cache = *package_cache;
+      pkgCache::PkgIterator pkg = cache.FindPkg (package);
+      if (!pkg.end())
+	mark_for_install (pkg);
+    }
+}
 
 /* Mark a package for removal and also remove as many of the packages
    that it depends on as possible.
@@ -993,7 +1118,9 @@ mark_for_remove (pkgCache::PkgIterator &pkg, bool only_maybe = false)
 {
   pkgDepCache &cache = *package_cache;
 
-  int n_broken = cache.BrokenCount ();
+  /* 'unsigned', so that we can handle more than 2^31 broken packages.
+   */
+  unsigned int n_broken = cache.BrokenCount ();
 
   cache.MarkDelete (pkg);
   cache[pkg].Flags &= ~pkgCache::Flag::Auto;
@@ -1028,6 +1155,9 @@ mark_for_remove (pkgCache::PkgIterator &pkg, bool only_maybe = false)
     }
 }
 
+/** COMMAND HANDLERS
+ */
+
 /* APTCMD_GET_PACKAGE_LIST 
 
    The get_package_list command can do some filtering and we have a
@@ -1137,6 +1267,7 @@ cmd_get_package_list ()
   bool only_installed = request.decode_int ();
   bool only_available = request.decode_int ();
   const char *pattern = request.decode_string_in_place ();
+  bool show_magic_sys = request.decode_int ();
 
   if (package_cache == NULL)
     {
@@ -1215,13 +1346,48 @@ cmd_get_package_list ()
       else
 	encode_empty_version_info (true);
 
-      // Available version and section
-      // XXX - avoid duplicating information
-      if (!candidate.end () && installed.CompareVer (candidate) < 0)
+      // Available version 
+      //
+      // We only offer an available version if the package is not
+      // installed at all, or if the available version is newer than
+      // the installed one, or if the installed version needs to be
+      // reinstalled and it is actually downloadable.
+
+      if (!candidate.end ()
+	  && (installed.end ()
+	      || installed.CompareVer (candidate) < 0
+	      || (pkg.State () == pkgCache::PkgIterator::NeedsUnpack
+		  && candidate.Downloadable ())))
 	encode_version_info (candidate, false);
       else
 	encode_empty_version_info (false);
     }
+
+  if (show_magic_sys)
+    {
+      // Append the "magic:sys" package that represents all system
+      // packages This artificial package is identified by its name and
+      // handled specially by MARK_NAMED_PACKAGE_FOR_INSTALL, etc.
+      
+      // Name
+      response.encode_string ("magic:sys");
+      
+      // Broken?  XXX - give real information here
+      response.encode_int (FALSE);
+
+      // Installed version
+      response.encode_string ("");
+      response.encode_int (1000);
+      response.encode_string ("system");
+      response.encode_string ("All system packages");
+      response.encode_string (NULL);
+      
+      // Available version
+      response.encode_string ("");
+      response.encode_string ("system");
+      response.encode_string (NULL);
+      response.encode_string (NULL);
+    }
 }
 
 /* APTCMD_GET_PACKAGE_INFO
@@ -1231,8 +1397,7 @@ cmd_get_package_list ()
  */
 
 static int
-installable_status_1 (pkgCache::PkgIterator &pkg,
-		      pkgCache::PkgIterator &want)
+installable_status_1 (pkgCache::PkgIterator &pkg)
 {
   pkgDepCache &Cache = *package_cache;
   pkgCache::VerIterator Ver = Cache[pkg].InstVerIter(Cache);
@@ -1275,7 +1440,7 @@ combine_status (int s1, int s2)
 }
 
 static int
-installable_status (pkgCache::PkgIterator &want)
+installable_status ()
 {
   pkgDepCache &cache = *package_cache;
   int installable_status = status_unable;
@@ -1286,15 +1451,14 @@ installable_status (pkgCache::PkgIterator &want)
     {
       if (cache[pkg].InstBroken())
 	installable_status =
-	  combine_status (installable_status_1 (pkg, want), 
-			  installable_status);
+	  combine_status (installable_status_1 (pkg), installable_status);
     }
 
   return installable_status;
 }
 
 static int
-removable_status (pkgCache::PkgIterator &want)
+removable_status ()
 {
   pkgDepCache &cache = *package_cache;
   for (pkgCache::PkgIterator pkg = cache.PkgBegin();
@@ -1327,31 +1491,32 @@ cmd_get_package_info ()
       pkgDepCache &cache = *package_cache;
       pkgCache::PkgIterator pkg = cache.FindPkg (package);
 
-      if (!pkg.end ())
+      // simulate install
+      
+      unsigned int old_broken_count = cache.BrokenCount();
+      mark_named_package_for_install (package);
+      if (cache.BrokenCount() > old_broken_count)
+	info.installable_status = installable_status ();
+      else
+	info.installable_status = status_able;
+      info.download_size = (int) cache.DebSize ();
+      info.install_user_size_delta = (int) cache.UsrSize ();
+      cache_reset ();
+
+      if (!only_installable_info)
 	{
-	  pkgCache::VerIterator inst = pkg.CurrentVer ();
-	  int old_broken_count;
-
-	  // simulate install
+	  // simulate remove
 	  
-	  old_broken_count = cache.BrokenCount();
-	  mark_for_install (pkg);
-	  if (cache.BrokenCount() > old_broken_count)
-	    info.installable_status = installable_status (pkg);
+	  if (!strcmp (package, "magic:sys"))
+	    {
+	      info.removable_status = status_unable;
+	    }
 	  else
-	    info.installable_status = status_able;
-	  info.download_size = (int) cache.DebSize ();
-	  info.install_user_size_delta = (int) cache.UsrSize ();
-	  cache_reset ();
-
-	  if (!only_installable_info)
 	    {
-	      // simulate remove
-	  
 	      old_broken_count = cache.BrokenCount();
 	      mark_for_remove (pkg);
 	      if (cache.BrokenCount() > old_broken_count)
-		info.removable_status = removable_status (pkg);
+		info.removable_status = removable_status ();
 	      else
 		info.removable_status = status_able;
 	      info.remove_user_size_delta = (int) cache.UsrSize ();
@@ -1412,7 +1577,8 @@ encode_dependencies (pkgCache::VerIterator &ver)
 }
 
 void
-encode_broken (pkgCache::PkgIterator &pkg, pkgCache::PkgIterator &want)
+encode_broken (pkgCache::PkgIterator &pkg,
+	       const char *want)
 {
   pkgDepCache &Cache = *package_cache;
   pkgCache::VerIterator Ver = Cache[pkg].InstVerIter(Cache);
@@ -1453,13 +1619,13 @@ encode_broken (pkgCache::PkgIterator &pkg, pkgCache::PkgIterator &want)
 	  /* Never blame conflicts on the package that we want to
 	     install.
 	  */
-	  if (target == want && Start->Type == pkgCache::Dep::Conflicts)
+	  if (target.Name() == want && Start->Type == pkgCache::Dep::Conflicts)
 	    g_string_append_printf (str, "%s", pkg.Name());
 	  else
 	    {
 	      g_string_append_printf (str, "%s", target.Name());
 	      if (Start.TargetVer() != 0)
-		g_string_append_printf (str, " %s %s",
+		g_string_append_printf (str, " (%s %s)",
 					Start.CompType(), Start.TargetVer());
 	    }
 
@@ -1487,7 +1653,7 @@ encode_package_and_version (const char *package, const char *version)
 }
 
 void
-encode_install_summary (pkgCache::PkgIterator &want)
+encode_install_summary (const char *want)
 {
   pkgDepCache &cache = *package_cache;
 
@@ -1498,7 +1664,7 @@ encode_install_summary (pkgCache::PkgIterator &want)
   if (cache.BrokenCount() > 0)
     fprintf (stderr, "[ Some installed packages are broken! ]\n");
 
-  mark_for_install (want);
+  mark_named_package_for_install (want);
 
   for (pkgCache::PkgIterator pkg = cache.PkgBegin();
        pkg.end() != true;
@@ -1544,8 +1710,6 @@ encode_remove_summary (pkgCache::PkgIterator &want)
        pkg.end() != true;
        pkg++)
     {
-      apt_proto_sumtype type;
-
       if (cache[pkg].Delete())
 	{
 	  response.encode_int (sumtype_removing);
@@ -1591,32 +1755,47 @@ cmd_get_package_details ()
   const char *version = request.decode_string_in_place ();
   int summary_kind = request.decode_int ();
 
-  pkgCache::PkgIterator pkg;
-  pkgCache::VerIterator ver;
-
-  if (find_package_version (package_cache, pkg, ver, package, version))
+  if (!strcmp (package, "magic:sys"))
     {
-      pkgDepCache &cache = *package_cache;
-      pkgRecords Recs (cache);
-      pkgRecords::Parser &P = Recs.Lookup (ver.FileList ());
-
-      response.encode_string (P.Maintainer().c_str());
-      response.encode_string (P.LongDesc().c_str());
-      encode_dependencies (ver);
+      response.encode_string ("");      // maintainer
+      response.encode_string 
+	("This is an artificial package that represents all\n"
+	 "system packages that are installed on your device.");
+      response.encode_int (deptype_end);  // dependencies
       if (summary_kind == 1)
-	encode_install_summary (pkg);
-      else if (summary_kind == 2)
-	encode_remove_summary (pkg);
+	encode_install_summary (package);
       else
 	response.encode_int (sumtype_end);
     }
   else
     {
-      // not found
-      response.encode_string (NULL);      // maintainer
-      response.encode_string (NULL);      // description
-      response.encode_int (deptype_end);  // dependencies
-      response.encode_int (sumtype_end);  // summary
+      pkgCache::PkgIterator pkg;
+      pkgCache::VerIterator ver;
+      
+      if (find_package_version (package_cache, pkg, ver, package, version))
+	{
+	  pkgDepCache &cache = *package_cache;
+	  pkgRecords Recs (cache);
+	  pkgRecords::Parser &P = Recs.Lookup (ver.FileList ());
+	  
+	  response.encode_string (P.Maintainer().c_str());
+	  response.encode_string (P.LongDesc().c_str());
+	  encode_dependencies (ver);
+	  if (summary_kind == 1)
+	    encode_install_summary (package);
+	  else if (summary_kind == 2)
+	    encode_remove_summary (pkg);
+	  else
+	    response.encode_int (sumtype_end);
+	}
+      else
+	{
+	  // not found
+	  response.encode_string (NULL);      // maintainer
+	  response.encode_string (NULL);      // description
+	  response.encode_int (deptype_end);  // dependencies
+	  response.encode_int (sumtype_end);  // summary
+	}
     }
 }
 
@@ -1678,7 +1857,6 @@ update_package_cache ()
       Fetcher.Clean (_config->FindDir("Dir::State::lists") + "partial/");
     }
   
-  // Prepare the cache.   
   cache_init ();
 
   return Failed? rescode_failure : rescode_success;
@@ -1778,15 +1956,9 @@ cmd_install_check ()
 
   if (package_cache)
     {
-      pkgDepCache &cache = *package_cache;
-      pkgCache::PkgIterator pkg = cache.FindPkg (package);
-
-      if (!pkg.end ())
-	{
-	  mark_for_install (pkg);
-	  result_code = operation (true);
-	  cache_reset ();
-	}
+      mark_named_package_for_install (package);
+      result_code = operation (true);
+      cache_reset ();
     }
 
   response.encode_int (result_code == rescode_success);
@@ -1814,14 +1986,8 @@ cmd_install_package ()
 
   if (package_cache)
     {
-      pkgDepCache &cache = *package_cache;
-      pkgCache::PkgIterator pkg = cache.FindPkg (package);
-
-      if (!pkg.end ())
-	{
-	  mark_for_install (pkg);
-	  result_code = operation (false);
-	}
+      mark_named_package_for_install (package);
+      result_code = operation (false);
     }
 
   need_cache_init ();
@@ -1988,11 +2154,12 @@ encode_upgrades ()
   response.encode_string (NULL);
 }
 
-/* We modify the pkgDPkgPM package manager slightly so that it never
-   attempts to configure packages that are not otherwise operated on
-   during a package management run.  We do this to so that we only get
-   failure reports for the actual packages being worked on, and not
-   for ones that failed to configure in an earlier run.
+/* We modify the pkgDPkgPM package manager so that we can provide our
+   own method of constructing the 'order list', the ordered list of
+   packages to handle.  We do this to ignore packages that should be
+   kept.  Ordinarily, if a package should be kept but needs to be
+   reinstalled or configured, libapt-pkg will try to do that with
+   every operation.
 
    XXX - there might be a way to get the same effect by cleverly
          manipulating the pkgDepCache, but I would have to look
@@ -2001,20 +2168,53 @@ encode_upgrades ()
 
 class myDPkgPM : public pkgDPkgPM
 {
-  virtual bool Configure(PkgIterator Pkg);
-  
 public:
 
+  bool CreateOrderList ();
+
   myDPkgPM(pkgDepCache *Cache);
 };
 
 bool
-myDPkgPM::Configure (PkgIterator Pkg)
+myDPkgPM::CreateOrderList ()
 {
-  for (int i = 0; i < List.size (); i++)
-    if (List[i].Pkg == Pkg)
-      return pkgDPkgPM::Configure (Pkg);
-  log_stderr ("not configuring unrelated package %s", Pkg.Name ());
+  if (pkgPackageManager::List != 0)
+    return true;
+   
+  delete pkgPackageManager::List;
+  pkgPackageManager::List = new pkgOrderList(&Cache);
+   
+  // Generate the list of affected packages and sort it
+  for (PkgIterator I = Cache.PkgBegin(); I.end() == false; I++)
+    {
+      // Ignore no-version packages
+      if (I->VersionList == 0)
+	continue;
+      
+      // Not interesting
+      if ((Cache[I].Keep() == true || 
+	   Cache[I].InstVerIter(Cache) == I.CurrentVer()) && 
+	  I.State() == pkgCache::PkgIterator::NeedsNothing &&
+	  (Cache[I].iFlags & pkgDepCache::ReInstall) != pkgDepCache::ReInstall &&
+	  (I.Purge() != false || Cache[I].Mode != pkgDepCache::ModeDelete ||
+	   (Cache[I].iFlags & pkgDepCache::Purge) != pkgDepCache::Purge))
+	continue;
+      
+      // Ignore interesting but kept packages, except when they are
+      // related to the current operation.
+
+      if (Cache[I].Keep() == true && !is_related (I))
+	{
+	  log_stderr ("Not handling unrelated package %s.", I.Name());
+	  continue;
+	}
+      
+      DBG ("Handling interesting package %s.", I.Name());
+
+      // Append it to the list
+      pkgPackageManager::List->push_back(I);      
+    }
+   
   return true;
 }
 
@@ -2040,17 +2240,15 @@ operation (bool check_only)
    pkgCacheFile &Cache = *package_cache;
 
    if (_config->FindB("APT::Get::Purge",false) == true)
-   {
-      pkgCache::PkgIterator I = Cache->PkgBegin();
-      for (; I.end() == false; I++)
-      {
-	 if (I.Purge() == false && Cache[I].Mode == pkgDepCache::ModeDelete)
-	    Cache->MarkDelete(I,true);
-      }
-   }
-   
-   bool Fail = false;
-   
+     {
+       pkgCache::PkgIterator I = Cache->PkgBegin();
+       for (; I.end() == false; I++)
+	 {
+	   if (I.Purge() == false && Cache[I].Mode == pkgDepCache::ModeDelete)
+	     Cache->MarkDelete(I,true);
+	 }
+     }
+
    if (Cache->DelCount() == 0 && Cache->InstCount() == 0 &&
        Cache->BadCount() == 0)
       return rescode_success;
@@ -2084,8 +2282,18 @@ operation (bool check_only)
        return rescode_failure;
      }
    
-   // Create the package manager and prepare to download
-   SPtr<pkgPackageManager> Pm = new myDPkgPM (Cache);
+   // Create the package manager
+   //
+   SPtr<myDPkgPM> Pm = new myDPkgPM (Cache);
+   
+   // Create the order list explicitely in a way that we like.  We
+   // have to do it explicitely since CreateOrderList is not virtual.
+   //
+   if (!Pm->CreateOrderList ())
+     return rescode_failure;
+
+   // Prepare to download
+   //
    if (Pm->GetArchives(&Fetcher,&List,&Recs) == false || 
        _error->PendingError() == true)
      return rescode_failure;
@@ -2227,16 +2435,54 @@ cmd_clean ()
     need_cache_init ();
 }
 
+static char *
+escape_for_shell (const char *str)
+{
+  char buf[2000];
+  char *p = buf;
+
+  /* Enclose the string in single quotes and escape single quotes.
+   */
+
+  *p++ = '\'';
+  while (*str)
+    {
+      if (p >= buf+sizeof(buf)-6)
+	return NULL;
+
+      if (*str == '\'')
+	{
+	  // Don't you love bourne shell syntax?
+	  *p++ = '\'';
+	  *p++ = '\\';
+	  *p++ = '\'';
+	  *p++ = '\'';
+	  str++;
+	}
+      else
+	*p++ = *str++;
+    }
+  *p++ = '\'';
+  *p++ = '\0';
+  return g_strdup (buf);
+}
+
 // XXX - interpret status codes
 
 static char *
 get_deb_record (const char *filename)
 {
-  char *cmd =
-    g_strdup_printf ("/usr/bin/dpkg-deb -f '%s'", filename);
+  char *esc_filename = escape_for_shell (filename);
+  if (esc_filename == NULL)
+    return NULL;
+
+  char *cmd = g_strdup_printf ("/usr/bin/dpkg-deb -f %s", esc_filename);
   fprintf (stderr, "%s\n", cmd);
   FILE *f = popen (cmd, "r");
+
   g_free (cmd);
+  g_free (esc_filename);
+
   if (f)
     {
       const size_t incr = 2000;
@@ -2551,10 +2797,17 @@ void
 cmd_install_file ()
 {
   const char *filename = request.decode_string_in_place ();
+  char *esc_filename = escape_for_shell (filename);
+  
+  if (esc_filename == NULL)
+    {
+      response.encode_int (0);
+      return;
+    }
 
   _system->UnLock();
 
-  char *cmd = g_strdup_printf ("/usr/bin/dpkg --install '%s'", filename);
+  char *cmd = g_strdup_printf ("/usr/bin/dpkg --install %s", esc_filename);
   fprintf (stderr, "%s\n", cmd);
   int res = system (cmd);
   g_free (cmd);
@@ -2563,13 +2816,15 @@ cmd_install_file ()
     {
       char *cmd =
 	g_strdup_printf ("/usr/bin/dpkg --purge "
-			 "`/usr/bin/dpkg-deb -f '%s' Package`",
-			 filename);
+			 "`/usr/bin/dpkg-deb -f %s Package`",
+			 esc_filename);
       fprintf (stderr, "%s\n", cmd);
       system (cmd);
       g_free (cmd);
     }
 
+  g_free (esc_filename);
+
   _system->Lock();
 
   need_cache_init ();
