@@ -1,14 +1,13 @@
 /*
  * This file is part of osso-application-installer
  *
- * Copyright (C) 2005 Nokia Corporation.
+ * Copyright (C) 2005, 2006 Nokia Corporation.  All Right reserved.
  *
  * Contact: Marius Vollmer <marius.vollmer@nokia.com>
  *
  * This program is free software; you can redistribute it and/or
- * modify it under the terms of the GNU General Public License as
- * published by the Free Software Foundation; either version 2 of the
- * License, or (at your option) any later version.
+ * modify it under the terms of the GNU General Public License version
+ * 2 as published by the Free Software Foundation.
  *
  * This program is distributed in the hope that it will be useful, but
  * WITHOUT ANY WARRANTY; without even the implied warranty of
@@ -27,11 +26,12 @@
 
 #include "main.h"
 
-void show_package_details (package_info *p, bool installed,
+void show_package_details (GtkWindow *parent,
+			   package_info *p, detail_kind kind,
 			   bool show_problems);
 
-char *decode_summary (apt_proto_decoder *dec,
-		      package_info *pi, bool installed);
+void decode_summary (apt_proto_decoder *dec,
+		     package_info *pi, detail_kind kind);
 void nicify_description_in_place (char *desc);
 
 #endif /* !DETAILS_H */
