@@ -1,14 +1,13 @@
 /*
  * This file is part of osso-application-installer
  *
- * Copyright (C) 2005, 2006 Nokia Corporation.
+ * Copyright (C) 2005, 2006 Nokia Corporation.  All Right reserved.
  *
  * Contact: Marius Vollmer <marius.vollmer@nokia.com>
  *
  * This program is free software; you can redistribute it and/or
- * modify it under the terms of the GNU General Public License as
- * published by the Free Software Foundation; either version 2 of the
- * License, or (at your option) any later version.
+ * modify it under the terms of the GNU General Public License version
+ * 2 as published by the Free Software Foundation.
  *
  * This program is distributed in the hope that it will be useful, but
  * WITHOUT ANY WARRANTY; without even the implied warranty of
@@ -25,6 +24,11 @@
 #ifndef INSTR_H
 #define INSTR_H
 
-void open_install_instructions (const char *filename);
+/* Carry out the instruction in FILENAME, which is assumed to refer to
+   a local file as produced by 'localize_file_and_keep_it_open'.
+   OPEN_LOCAL_INSTALL_INSTRUCTIONS will call 'cleanup_temp_file' when
+   it is done with FILENAME.
+*/
+void open_local_install_instructions (const char *filename);
 
 #endif /* !INSTR_H */
