@@ -4,14 +4,13 @@
  * Parts of this file are derived from apt.  Apt is copyright 1997,
  * 1998, 1999 Jason Gunthorpe and others.
  *
- * Copyright (C) 2005 Nokia Corporation.
+ * Copyright (C) 2005, 2006 Nokia Corporation.  All Right reserved.
  *
  * Contact: Marius Vollmer <marius.vollmer@nokia.com>
  *
  * This program is free software; you can redistribute it and/or
- * modify it under the terms of the GNU General Public License as
- * published by the Free Software Foundation; either version 2 of the
- * License, or (at your option) any later version.
+ * modify it under the terms of the GNU General Public License version
+ * 2 as published by the Free Software Foundation.
  *
  * This program is distributed in the hope that it will be useful, but
  * WITHOUT ANY WARRANTY; without even the implied warranty of
@@ -44,6 +43,7 @@
 #include "settings.h"
 #include "search.h"
 #include "instr.h"
+#include "repo.h"
 
 #define MAX_PACKAGES_NO_CATEGORIES 7
 
@@ -61,7 +61,7 @@ using namespace std;
 static guint apt_source_id;
 
 static void set_details_callback (void (*func) (gpointer), gpointer data);
-static void set_operation_label (const char *label);
+static void set_operation_label (const char *label, const char *insens);
 static void set_operation_callback (void (*func) (gpointer), gpointer data);
 static void enable_search (bool f);
 static void set_current_help_topic (const char *topic);
@@ -74,11 +74,12 @@ struct view {
   GtkWidget *(*maker) (view *);
 };
 
-GtkWidget *main_vbox;
-GtkWidget *main_trail;
+GtkWidget *main_vbox = NULL;
+GtkWidget *main_trail = NULL;
+GtkWidget *device_label = NULL;
 GtkWidget *cur_view = NULL;
-view *cur_view_struct;
-GList *cur_path;
+view *cur_view_struct = NULL;
+GList *cur_path = NULL;
 
 static GList *
 make_view_path (view *v)
@@ -93,12 +94,17 @@ void
 show_view (view *v)
 {
   if (cur_view)
-    gtk_widget_destroy (cur_view);
+    {
+      gtk_container_remove(GTK_CONTAINER(main_vbox), cur_view);
+      cur_view = NULL;
+    }
 
   set_details_callback (NULL, NULL);
-  set_operation_label (NULL);
+  set_operation_label (NULL, NULL);
   set_operation_callback (NULL, NULL);
 
+  allow_updating ();
+
   cur_view = v->maker (v);
   cur_view_struct = v;
 
@@ -107,6 +113,7 @@ show_view (view *v)
   hildon_bread_crumb_trail_set_path (main_trail, cur_path);
   
   gtk_box_pack_start (GTK_BOX (main_vbox), cur_view, TRUE, TRUE, 10);
+  gtk_widget_show(main_vbox);
 }
 
 static void
@@ -198,38 +205,78 @@ make_padded_button (const char *label,
   return btn;
 }
 
-static GdkPixbuf *main_image = NULL;
+static gboolean
+expose_main_view (GtkWidget *w, GdkEventExpose *ev, gpointer data)
+{
+  /* This puts the background pixmap for the ACTIVE state into the
+     lower right corner.  Using bg_pixmap[ACTIVE] is a hack to
+     communicate which pixmap to use from the gtkrc file.  The widget
+     will never actually be in the ACTIVE state.
+  */
+
+  GtkStyle *style = gtk_rc_get_style (w);
+  GdkPixmap *pixmap = style->bg_pixmap[GTK_STATE_ACTIVE];
+  gint ww, wh, pw, ph;
+
+  if (pixmap)
+    {
+      gdk_drawable_get_size (pixmap, &pw, &ph);
+      gdk_drawable_get_size (w->window, &ww, &wh);
+      
+      gdk_draw_drawable (w->window, style->fg_gc[GTK_STATE_NORMAL],
+			 pixmap, 0, 0, ww-pw, wh-ph, pw, ph);
+    }
+
+  gtk_container_propagate_expose (GTK_CONTAINER (w),
+				  gtk_bin_get_child (GTK_BIN (w)),
+				  ev);
+
+  return TRUE;
+}
+
+static void
+device_label_destroyed (GtkWidget *widget, gpointer data)
+{
+  if (device_label == widget)
+    device_label = NULL;
+}
 
 GtkWidget *
 make_main_view (view *v)
 {
   GtkWidget *view;
-  GtkWidget *table, *vbox;
+  GtkWidget *vbox, *hbox;
   GtkWidget *btn, *label, *image;
+  GtkSizeGroup *btn_group;
 
-  view = gtk_hbox_new (FALSE, 0);
-  vbox = gtk_vbox_new (FALSE, 0);
-  table = gtk_table_new (5, 2, FALSE);
-  gtk_box_pack_start (GTK_BOX (vbox), table, FALSE, FALSE, 0);
-  gtk_box_pack_start (GTK_BOX (view), vbox, FALSE, FALSE, 0);
+  btn_group = gtk_size_group_new(GTK_SIZE_GROUP_BOTH);
 
-  gtk_table_set_row_spacings (GTK_TABLE (table), 10);
-  gtk_table_set_col_spacings (GTK_TABLE (table), 10);
+  view = gtk_event_box_new ();
+  gtk_widget_set_name (view, "osso-application-installer-main-view");
 
+  g_signal_connect (view, "expose-event",
+                    G_CALLBACK (expose_main_view), NULL);
+
+  vbox = gtk_vbox_new (FALSE, 10);
+  gtk_container_add (GTK_CONTAINER (view), vbox);
+
+  // first label
+  hbox = gtk_hbox_new (FALSE, 10);
   image = gtk_image_new_from_icon_name ("qgn_list_filesys_divc_cls",
 					HILDON_ICON_SIZE_26);
-  gtk_table_attach_defaults (GTK_TABLE (table), image,
-			     0, 1, 0, 1);
+  gtk_box_pack_start (GTK_BOX (hbox), image, FALSE, FALSE, 0);
 
-  // XXX - The usize request below prevents the label from expanding
-  //       the table column if the device is too long.
-
-  label = gtk_label_new (device_name ());
-  gtk_widget_set_usize (label, 10, -1);
-  gtk_misc_set_alignment (GTK_MISC (label), 0.0, 0.5);
-  gtk_table_attach_defaults (GTK_TABLE (table), label,
-			     1, 2, 0, 1);
+  device_label = gtk_label_new (device_name ());
+  gtk_label_set_ellipsize(GTK_LABEL (device_label), PANGO_ELLIPSIZE_END);
+  gtk_misc_set_alignment (GTK_MISC (device_label), 0.0, 0.5);
+  gtk_box_pack_start (GTK_BOX (hbox),  device_label, TRUE, TRUE, 0);
+  gtk_box_pack_start (GTK_BOX (vbox), hbox,  FALSE, FALSE, 0);
   
+  g_signal_connect (device_label, "destroy",
+		    G_CALLBACK (device_label_destroyed), NULL);
+
+  // first button
+  hbox = gtk_hbox_new (FALSE, 0);
   btn = make_padded_button (_("ai_li_uninstall"),
 			    OssoGtkButtonAttachFlags
 			    (OSSO_GTK_BUTTON_ATTACH_NORTH |
@@ -239,19 +286,25 @@ make_main_view (view *v)
   g_signal_connect (G_OBJECT (btn), "clicked",
 		    G_CALLBACK (show_view_callback),
 		    &uninstall_applications_view);
-  gtk_table_attach_defaults (GTK_TABLE (table), btn,
-			     1, 2, 1, 2);
+  gtk_size_group_add_widget(btn_group, btn);
+  // 36 padding = 26 icon size + 10 padding
+  gtk_box_pack_start (GTK_BOX (hbox),  btn,  FALSE, FALSE, 36); 
+  gtk_box_pack_start (GTK_BOX (vbox), hbox, FALSE, FALSE, 0);
   grab_focus_on_map (btn);
 
+  // second label
+  hbox = gtk_hbox_new (FALSE, 10);
   image = gtk_image_new_from_icon_name ("qgn_list_browser",
 					HILDON_ICON_SIZE_26);
-  gtk_table_attach_defaults (GTK_TABLE (table), image,
-			     0, 1, 2, 3);
+  gtk_box_pack_start (GTK_BOX (hbox), image, FALSE, FALSE, 0);
+
   label = gtk_label_new (_("ai_li_repository"));
   gtk_misc_set_alignment (GTK_MISC (label), 0.0, 0.5);
-  gtk_table_attach_defaults (GTK_TABLE (table), label,
-			     1, 2, 2, 3);
+  gtk_box_pack_start (GTK_BOX (hbox),  label, FALSE, FALSE, 0);
+  gtk_box_pack_start (GTK_BOX (vbox), hbox,  FALSE, FALSE, 0);
 
+  // second button
+  hbox = gtk_hbox_new (FALSE, 0);
   btn = make_padded_button (_("ai_li_install"),
 			    OssoGtkButtonAttachFlags
 			    (OSSO_GTK_BUTTON_ATTACH_NORTH |
@@ -260,9 +313,13 @@ make_main_view (view *v)
   g_signal_connect (G_OBJECT (btn), "clicked",
 		    G_CALLBACK (show_view_callback),
 		    &install_applications_view);
-  gtk_table_attach_defaults (GTK_TABLE (table), btn,
-			     1, 2, 3, 4);
+  gtk_size_group_add_widget(btn_group, btn);
+  // 36 padding = 26 icon size + 10 padding
+  gtk_box_pack_start (GTK_BOX (hbox),  btn,  FALSE, FALSE, 36);
+  gtk_box_pack_start (GTK_BOX (vbox), hbox, FALSE, FALSE, 0);
   
+  // third button
+  hbox = gtk_hbox_new (FALSE, 0);
   btn = make_padded_button (_("ai_li_update"),
 			    OssoGtkButtonAttachFlags
 			    (OSSO_GTK_BUTTON_ATTACH_EAST |
@@ -271,42 +328,21 @@ make_main_view (view *v)
   g_signal_connect (G_OBJECT (btn), "clicked",
 		    G_CALLBACK (show_view_callback),
 		    &upgrade_applications_view);
-  gtk_table_attach_defaults (GTK_TABLE (table), btn,
-			     1, 2, 4, 5);
+  gtk_size_group_add_widget(btn_group, btn);
+  // 36 padding = 26 icon size + 10 padding
+  gtk_box_pack_start (GTK_BOX (hbox),  btn,  FALSE, FALSE, 36);
+  gtk_box_pack_start (GTK_BOX (vbox), hbox, FALSE, FALSE, 0);
 
-  if (red_pill_mode)
-    {
-      btn = make_padded_button (_("ai_me_tools_refresh"),
-				OssoGtkButtonAttachFlags
-				(OSSO_GTK_BUTTON_ATTACH_NORTH |
-				 OSSO_GTK_BUTTON_ATTACH_EAST |
-				 OSSO_GTK_BUTTON_ATTACH_SOUTH |
-				 OSSO_GTK_BUTTON_ATTACH_WEST));
-      g_signal_connect (G_OBJECT (btn), "clicked",
-			G_CALLBACK (refresh_package_cache),
-			NULL);
-      gtk_table_attach_defaults (GTK_TABLE (table), btn,
-				 1, 2, 5, 6);
-    }
-
-  if (main_image == NULL)
-    {
-      // XXX - find a proper way to theme the main view.
-      main_image = gdk_pixbuf_new_from_file ("/usr/share/themes/default/images/qgn_plat_application_installer_image.png", NULL);
-    }
-
-  vbox = gtk_vbox_new (FALSE, 0);
-  gtk_box_pack_end (GTK_BOX (view), vbox, FALSE, FALSE, 0);
-  image = gtk_image_new_from_pixbuf (main_image);
-  gtk_box_pack_end (GTK_BOX (vbox), image, FALSE, FALSE, 0);
-  
   gtk_widget_show_all (view);
+  g_object_unref(btn_group);
 
   get_package_list_info (NULL);
 
   enable_search (false);
   set_current_help_topic (AI_TOPIC ("mainview"));
 
+  prevent_updating ();
+
   return view;
 }
 
@@ -332,10 +368,12 @@ package_info::package_info ()
 
   have_info = false;
 
-  have_details = false;
+  have_detail_kind = no_details;
   maintainer = NULL;
   description = NULL;
   summary = NULL;
+  for (int i = 0; i < sumtype_max; i++)
+    summary_packages[i] = NULL;
   dependencies = NULL;
 
   model = NULL;
@@ -359,6 +397,11 @@ package_info::~package_info ()
   g_free (maintainer);
   g_free (description);
   g_free (summary);
+  for (int i = 0; i < sumtype_max; i++)
+    {
+      g_list_foreach (summary_packages[i], (GFunc) g_free, NULL);
+      g_list_free (summary_packages[i]);
+    }
   g_free (dependencies);
   g_free (filename);
 }
@@ -452,7 +495,7 @@ free_all_packages ()
 static const char *
 canonicalize_section_name (const char *name)
 {
-  if (name == NULL || red_pill_mode)
+  if (name == NULL || (red_pill_mode && red_pill_show_all))
     return name;
 
   if (g_str_has_prefix (name, "maemo/"))
@@ -467,7 +510,7 @@ canonicalize_section_name (const char *name)
 const char *
 nicify_section_name (const char *name)
 {
-  if (red_pill_mode)
+  if (red_pill_mode && red_pill_show_all)
     return name;
 
   name = canonicalize_section_name (name);
@@ -651,7 +694,7 @@ get_package_list_reply (int cmd, apt_proto_decoder *dec, void *data)
   if (dec == NULL)
     ;
   else if (dec->decode_int () == 0)
-    annoy_user_with_log ("Operation failed");
+    annoy_user_with_log (_("ai_ni_operation_failed"));
   else
     {
       section_info *all_si = new section_info;
@@ -675,9 +718,9 @@ get_package_list_reply (int cmd, apt_proto_decoder *dec, void *data)
 	  info->available_short_description = dec->decode_string_dup ();
 	  available_icon = dec->decode_string_in_place ();
 
-	  info->installed_icon = pixbuf_from_base64 (installed_icon);
-	  if (available_icon)
-	    info->available_icon = pixbuf_from_base64 (available_icon);
+          info->installed_icon = pixbuf_from_base64 (installed_icon);
+          if (available_icon)
+            info->available_icon = pixbuf_from_base64 (available_icon);
 	  else
 	    {
 	      info->available_icon = info->installed_icon;
@@ -764,10 +807,11 @@ get_package_list_with_cont (void (*cont) (void *data), void *data)
   free_all_packages ();
 
   show_updating ();
-  apt_worker_get_package_list (!red_pill_mode,
+  apt_worker_get_package_list (!(red_pill_mode && red_pill_show_all),
 			       false, 
 			       false, 
 			       NULL,
+			       red_pill_mode && red_pill_show_magic_sys,
 			       get_package_list_reply, c);
 }
 
@@ -896,7 +940,10 @@ get_intermediate_package_info (package_info *pi,
       intermediate_data = data;
     }
   else
-    printf ("package info request already pending.\n");
+    {
+      printf ("package info request already pending.\n");
+      pi->unref();
+    }
 
   if (cur_packages_for_info == NULL && old_intermediate_info == NULL)
     get_next_package_info (NULL, NULL, false);
@@ -934,7 +981,7 @@ struct rpc_clos {
 };
 
 static void
-refresh_package_cache_cont (void *data)
+refresh_package_cache_cont2 (void *data)
 {
   rpc_clos *c = (rpc_clos *)data;
   if (c->cont)
@@ -981,7 +1028,12 @@ refresh_package_cache_reply (int cmd, apt_proto_decoder *dec, void *data)
     {
       int result_code = dec->decode_int ();
 
-      if (result_code != rescode_success)
+      if (result_code == rescode_download_failed)
+	{
+	  annoy_user (_("ai_ni_error_download_failed"));
+	  success = false;
+	}
+      else if (result_code != rescode_success)
 	{
 	  annoy_user (_("ai_ni_update_list_not_successful"));
 	  success = false;
@@ -1000,7 +1052,7 @@ refresh_package_cache_reply (int cmd, apt_proto_decoder *dec, void *data)
   // failed because it might have partially succeeded and changed
   // anyway.  So we need to resynchronize.
   
-  get_package_list_with_cont (refresh_package_cache_cont, c);
+  get_package_list_with_cont (refresh_package_cache_cont2, c);
 }
 
 static bool refreshed_this_session = false;
@@ -1021,7 +1073,8 @@ refresh_package_cache_cont (bool res, void *data)
 }
 
 void
-refresh_package_cache_with_cont (void (*cont) (bool res, void *data), 
+refresh_package_cache_with_cont (bool ask,
+				 void (*cont) (bool res, void *data), 
 				 void *data)
 {
   refreshed_this_session = true;
@@ -1030,13 +1083,16 @@ refresh_package_cache_with_cont (void (*cont) (bool res, void *data),
   c->cont = cont;
   c->data = data;
 
-  ask_yes_no (_("ai_nc_confirm_update"), refresh_package_cache_cont, c);
+  if (ask)
+    ask_yes_no (_("ai_nc_confirm_update"), refresh_package_cache_cont, c);
+  else
+    refresh_package_cache_cont (true, c);
 }
 
 void
-refresh_package_cache ()
+refresh_package_cache (bool ask)
 {
-  refresh_package_cache_with_cont (NULL, NULL);
+  refresh_package_cache_with_cont (ask, NULL, NULL);
 }
 
 static int
@@ -1070,7 +1126,7 @@ maybe_refresh_package_cache ()
       && days_elapsed_since (last_update) < 30)
     return;
 
-  refresh_package_cache ();
+  refresh_package_cache (true);
 }
 
 static bool
@@ -1091,7 +1147,7 @@ confirm_install (package_info *pi,
   ask_yes_no_with_details ((pi->installed_version
 			    ? _("ai_ti_confirm_update")
 			    : _("ai_ti_confirm_install")),
-			   text->str, pi, false, cont, data);
+			   text->str, pi, install_details, cont, data);
   g_string_free (text, 1);
 }
 
@@ -1117,7 +1173,8 @@ install_package_reply (int cmd, apt_proto_decoder *dec, void *data)
       return;
     }
 
-  int result_code = dec->decode_int ();
+  apt_proto_result_code result_code =
+    apt_proto_result_code (dec->decode_int ());
 
   if (clean_after_install)
     apt_worker_clean (clean_reply, NULL);
@@ -1136,9 +1193,12 @@ install_package_reply (int cmd, apt_proto_decoder *dec, void *data)
   else
     {
       if (progress_was_cancelled ())
-	annoy_user ((upgrading
-		     ? _("ai_ni_update_cancelled")
-		     : _("ai_ni_install_cancelled")));
+	{
+	  if (ui_version < 2)
+	    annoy_user ((upgrading
+			 ? _("ai_ni_update_cancelled")
+			 : _("ai_ni_install_cancelled")));
+	}
       else
 	{
 	  char *str =
@@ -1146,6 +1206,9 @@ install_package_reply (int cmd, apt_proto_decoder *dec, void *data)
 			      ? _("ai_ni_error_update_failed")
 			      : _("ai_ni_error_installation_failed")),
 			     pi->name);
+
+	  result_code = scan_log_for_result_code (result_code);
+
 	  annoy_user_with_result_code (result_code, str, upgrading);
 	  g_free (str);
 	}
@@ -1193,6 +1256,7 @@ install_package_cont3 (bool res, void *data)
     {
       if (res)
 	{
+	  set_log_start ();
 	  apt_worker_install_package (c->pi->name,
 				      c->pi->installed_version != NULL,
 				      install_package_reply, c->pi);
@@ -1234,9 +1298,10 @@ install_package_cont5 (bool res, void *data)
 
   if (!res)
     {
-      annoy_user (c->pi->installed_version
-		  ? _("ai_ni_update_cancelled")
-		  : _("ai_ni_install_cancelled"));
+      if (ui_version < 2)
+	annoy_user (c->pi->installed_version
+		    ? _("ai_ni_update_cancelled")
+		    : _("ai_ni_install_cancelled"));
       install_package_cont3 (false, data);
     }
   else
@@ -1320,14 +1385,14 @@ annoy_user_with_installable_status_details (package_info *pi)
       annoy_user_with_details ((pi->installed_version
 				? _("ai_ni_error_update_missing")
 				: _("ai_ni_error_install_missing")),
-			       pi, false);
+			       pi, install_details);
     }
   else if (pi->info.installable_status == status_conflicting)
     {
       annoy_user_with_details ((pi->installed_version
 				? _("ai_ni_error_update_conflict")
 				: _("ai_ni_error_install_conflict")),
-			       pi, false);
+			       pi, install_details);
     }
   else if (pi->info.installable_status == status_corrupted)
     {
@@ -1352,7 +1417,7 @@ annoy_user_with_installable_status_details (package_info *pi)
 			  ? _("ai_ni_error_update_failed")
 			  : _("ai_ni_error_installation_failed")),
 			 pi->name);
-      annoy_user_with_details (str, pi, false);
+      annoy_user_with_details (str, pi, install_details);
       g_free (str);
     }
 }
@@ -1401,10 +1466,13 @@ install_package_cont2 (bool res, void *data)
     }
   else
     {
-      if (pi->installed_version)
-	annoy_user (_("ai_ni_update_cancelled"));
-      else
-	annoy_user (_("ai_ni_install_cancelled"));
+      if (ui_version < 2)
+	{
+	  if (pi->installed_version)
+	    annoy_user (_("ai_ni_update_cancelled"));
+	  else
+	    annoy_user (_("ai_ni_install_cancelled"));
+	}
       pi->unref ();
     }
 }
@@ -1426,7 +1494,7 @@ static void
 available_package_details (gpointer data)
 {
   package_info *pi = (package_info *)data;
-  show_package_details (pi, false, false);
+  show_package_details (get_main_window (), pi, install_details, false);
 }
 
 void
@@ -1449,7 +1517,7 @@ static void
 installed_package_details (gpointer data)
 {
   package_info *pi = (package_info *)data;
-  show_package_details (pi, true, false);
+  show_package_details (get_main_window (), pi, remove_details, false);
 }
 
 static void uninstall_package (package_info *);
@@ -1472,11 +1540,22 @@ installed_package_selected (package_info *pi)
 static GtkWidget *
 make_last_update_label ()
 {
-  char text[1024];
-  time_t t = last_update;
-  struct tm *broken = localtime (&t);
-  strftime (text, 1024, _("ai_li_updated_%x"), broken);
-  return gtk_label_new (text);
+  char time_string[1024];
+
+  if (last_update == 0)
+    strncpy (time_string, _("ai_li_never"), 1024);
+  else
+    {
+      time_t t = last_update;
+      struct tm *broken = localtime (&t);
+      strftime (time_string, 1024, "%x", broken);
+    }
+
+  char *text = g_strdup_printf (_("ai_li_updated_%s"), time_string);
+  GtkWidget *label = gtk_label_new (text);
+  g_free (text);
+
+  return label;
 }
 
 static GtkWidget *
@@ -1499,20 +1578,13 @@ make_package_list_view (GtkWidget *list_widget,
   return view;
 }
 
-static void
-set_install_section_name (const char *name)
-{
-  g_free ((gchar *)install_section_view.label);
-  install_section_view.label = g_strdup (nicify_section_name (name));
-}
-
 GtkWidget *
 make_install_section_view (view *v)
 {
   GtkWidget *view;
-  GList *packages;
 
-  set_operation_label (_("ai_me_package_install"));
+  set_operation_label (_("ai_me_package_install"),
+		       _("ai_ib_nothing_to_install"));
 
   section_info *si = find_section_info (&install_sections,
 					cur_section_name, false, true);
@@ -1548,14 +1620,52 @@ view_section (section_info *si)
   show_view (&install_section_view);
 }
 
+static void
+check_repositories_reply (int cmd, apt_proto_decoder *dec, void *data)
+{
+  if (dec == NULL)
+    return;
+
+  while (!dec->corrupted ())
+    {
+      const char *line = dec->decode_string_in_place ();
+      char *start, *end;
+
+      if (line == NULL)
+	break;
+
+      start = (char *)line;
+      if (parse_quoted_word (&start, &end, false)
+	  && end - start == 3 && !strncmp (start, "deb", 3))
+	{
+	  /* Found one active repository.  Maybe we need to refresh
+	     it.
+	  */
+	  maybe_refresh_package_cache ();
+	  return;
+	}
+    }
+
+  /* No repositories active.
+   */
+  irritate_user (_("ai_ib_no_repositories"));
+}
+
+static void
+check_repositories ()
+{
+  apt_worker_get_sources_list (check_repositories_reply, NULL);
+}
+
 GtkWidget *
 make_install_applications_view (view *v)
 {
   GtkWidget *list, *view, *label;
 
-  maybe_refresh_package_cache ();
+  check_repositories ();
 
-  set_operation_label (_("ai_me_package_install"));
+  set_operation_label (_("ai_me_package_install"),
+		       _("ai_ib_nothing_to_install"));
 
   if (install_sections && install_sections->next == NULL)
     {
@@ -1597,9 +1707,10 @@ make_upgrade_applications_view (view *v)
 {
   GtkWidget *view;
 
-  maybe_refresh_package_cache ();
+  check_repositories ();
 
-  set_operation_label (_("ai_me_package_update"));
+  set_operation_label (_("ai_me_package_update"),
+		       _("ai_ib_nothing_to_update"));
 
   GtkWidget *list =
     make_global_package_list (upgradeable_packages,
@@ -1632,7 +1743,7 @@ confirm_uninstall (package_info *pi,
 		   pi->name, pi->installed_version, size_buf);
 
   ask_yes_no_with_details (_("ai_ti_confirm_uninstall"), text->str,
-			   pi, true, cont, data);
+			   pi, remove_details, cont, data);
   g_string_free (text, 1);
 }
 
@@ -1675,12 +1786,12 @@ annoy_user_with_removable_status_details (package_info *pi)
 {
   if (pi->info.removable_status == status_needed)
     annoy_user_with_details (_("ai_ni_error_uninstall_packagesneeded"),
-			     pi, true);
+			     pi, remove_details);
   else
     {
       char *str = g_strdup_printf (_("ai_ni_error_uninstallation_failed"),
 				   pi->name);
-      annoy_user_with_details (str, pi, true);
+      annoy_user_with_details (str, pi, remove_details);
       g_free (str);
     }
 }
@@ -1797,7 +1908,8 @@ uninstall_package_cont2 (bool res, void *data)
     check_uninstall_scripts (pi);
   else
     {
-      annoy_user (_("ai_ni_uninstall_cancelled"));
+      if (ui_version < 2)
+	annoy_user (_("ai_ni_uninstall_cancelled"));
       pi->unref ();
     }
 }
@@ -1819,9 +1931,9 @@ GtkWidget *
 make_uninstall_applications_view (view *v)
 {
   GtkWidget *view;
-  GList *packages;
 
-  set_operation_label (_("ai_me_package_uninstall"));
+  set_operation_label (_("ai_me_package_uninstall"),
+		       _("ai_ib_nothing_to_uninstall"));
 
   view = make_global_package_list (installed_packages,
 				   true,
@@ -1847,7 +1959,10 @@ make_search_results_view (view *v)
     {
       set_operation_label (v->parent == &install_applications_view
 			   ? _("ai_me_package_install")
-			   : _("ai_me_package_update"));
+			   : _("ai_me_package_update"),
+			   v->parent == &install_applications_view
+			   ? _("ai_ib_nothing_to_install")
+			   : _("ai_ib_nothing_to_update"));
 
       view = make_global_package_list (search_result_packages,
 				       false,
@@ -1861,7 +1976,8 @@ make_search_results_view (view *v)
     }
   else
     {
-      set_operation_label (_("ai_me_package_uninstall"));
+      set_operation_label (_("ai_me_package_uninstall"),
+			   _("ai_ib_nothing_to_uninstall"));
 
       view = make_global_package_list (search_result_packages,
 				       true,
@@ -1897,19 +2013,6 @@ search_package_list (GList **result,
     }
 }
 
-static void
-search_section_list (GList **result,
-		     GList *sections, const char *pattern)
-{
-  while (sections)
-    {
-      section_info *si = (section_info *)sections->data;
-      search_package_list (result, si->packages, pattern);
-
-      sections = sections->next;
-    }
-}
-
 static void
 find_in_package_list (GList **result,
 		      GList *packages, const char *name)
@@ -1946,6 +2049,8 @@ search_packages_reply (int cmd, apt_proto_decoder *dec, void *data)
 {
   view *parent = (view *)data;
 
+  hide_updating ();
+
   if (dec == NULL)
     return;
 
@@ -2050,14 +2155,17 @@ search_packages (const char *pattern, bool in_descriptions)
     }
   else
     {
+      show_updating (_("ai_nw_searching"));
+
       bool only_installed = (parent == &uninstall_applications_view
 			     || parent == &upgrade_applications_view);
       bool only_available = (parent == &install_applications_view
 			     || parent == &upgrade_applications_view);
-      apt_worker_get_package_list (!red_pill_mode,
+      apt_worker_get_package_list (!(red_pill_mode && red_pill_show_all),
 				   only_installed,
 				   only_available, 
 				   pattern,
+				   red_pill_mode && red_pill_show_magic_sys,
 				   search_packages_reply, parent);
     }
 }
@@ -2072,12 +2180,17 @@ install_named_package (const char *package)
   find_in_package_list (&p, installed_packages, package);
 
   if (p == NULL)
-    annoy_user ("Not found?");
+    annoy_user (_("ai_ni_error_download_missing"));
   else
     {
       package_info *pi = (package_info *)p->data;
       if (pi->available_version == NULL)
-	annoy_user ("Already latest version.");
+	{
+	  char *text = g_strdup_printf (_("ai_ni_package_installed"),
+					package);
+	  annoy_user (text);
+	  g_free (text);
+	}
       else
 	install_package (pi);
     }
@@ -2106,6 +2219,7 @@ set_details_callback (void (*func) (gpointer), gpointer data)
 static void (*operation_func) (gpointer);
 static gpointer operation_data;
 static const char *operation_label;
+static const char *insensitive_operation_press_label = NULL;
 
 void
 do_current_operation ()
@@ -2117,13 +2231,20 @@ do_current_operation ()
 static void set_operation_toolbar_label (const char *label, bool enable);
 
 static void
-set_operation_label (const char *label)
+set_operation_label (const char *label, const char *insens)
 {
   if (label == NULL)
-    label = _("ai_me_package_install");
+    {
+      label = _("ai_me_package_install");
+      insens = _("ai_ib_nothing_to_install");
+    }
+
+  if (ui_version < 2 || insens == NULL)
+    insens = _("ai_ib_not_available");
 
   operation_label = label;
-  set_operation_menu_label (label, operation_func != NULL);
+  insensitive_operation_press_label = insens;
+  set_operation_menu_label (label, operation_func != NULL, insens);
   set_operation_toolbar_label (label, operation_func != NULL);
 }
 
@@ -2132,7 +2253,8 @@ set_operation_callback (void (*func) (gpointer), gpointer data)
 {
   operation_data = data;
   operation_func = func;
-  set_operation_label (operation_label);
+  set_operation_label (operation_label,
+		       insensitive_operation_press_label);
 }
 
 static void
@@ -2168,7 +2290,13 @@ install_from_file_reply (int cmd, apt_proto_decoder *dec, void *data)
 				   ? _("ai_ni_error_update_failed")
 				   : _("ai_ni_error_installation_failed"),
 				   pi->name);
-      annoy_user_with_log (str);
+
+      apt_proto_result_code result_code = rescode_failure;
+
+      result_code = scan_log_for_result_code (result_code);
+
+      annoy_user_with_result_code (result_code, str,
+				   pi->installed_version != NULL);
       g_free (str);
     }
 
@@ -2185,13 +2313,15 @@ install_from_file_cont4 (bool res, void *data)
       show_progress (pi->installed_version
 		     ? _("ai_nw_updating")
 		     : _("ai_nw_installing"));
+      set_log_start ();
       apt_worker_install_file (pi->filename,
 			       install_from_file_reply, pi);
     }
   else
     {
       cleanup_temp_file ();
-      annoy_user (_("ai_ni_install_cancelled"));
+      if (ui_version < 2)
+	annoy_user (_("ai_ni_install_cancelled"));
       pi->unref ();
     }
 }
@@ -2263,7 +2393,7 @@ file_details_reply (int cmd, apt_proto_decoder *dec, void *data)
   pi->available_icon = pixbuf_from_base64 (dec->decode_string_in_place ());
 
   pi->have_info = true;
-  pi->have_details = true;
+  pi->have_detail_kind = install_details;
 
   
   if (pi->info.installable_status == status_incompatible)
@@ -2273,7 +2403,7 @@ file_details_reply (int cmd, apt_proto_decoder *dec, void *data)
   else if (pi->info.installable_status == status_corrupted)
     pi->summary = g_strdup (_("ai_ni_error_install_corrupted"));
   else
-    pi->summary = decode_summary (dec, pi, false);
+    decode_summary (dec, pi, install_details);
 
   GString *text = g_string_new ("");
 
@@ -2300,7 +2430,7 @@ file_details_reply (int cmd, apt_proto_decoder *dec, void *data)
 			    ? _("ai_ti_confirm_update")
 			    : _("ai_ti_confirm_install")),
 			   text->str,
-			   pi, false, cont, pi);
+			   pi, install_details, cont, pi);
     
   g_string_free (text, 1);
 }
@@ -2312,11 +2442,12 @@ install_from_file_cont2 (char *filename, void *unused)
     {
       if (g_str_has_suffix (filename, ".install"))
 	{
-	  open_install_instructions (filename);
+	  open_local_install_instructions (filename);
 	  g_free (filename);
 	}
       else
-	apt_worker_get_file_details (!red_pill_mode, filename,
+	apt_worker_get_file_details (!(red_pill_mode && red_pill_show_all),
+				     filename,
 				     file_details_reply, filename);
     }
 }
@@ -2324,7 +2455,7 @@ install_from_file_cont2 (char *filename, void *unused)
 static void
 install_from_file_cont (char *uri, void *unused)
 {
-  localize_file (uri, install_from_file_cont2, NULL);
+  localize_file_and_keep_it_open (uri, install_from_file_cont2, NULL);
   g_free (uri);
 }
 
@@ -2340,13 +2471,25 @@ window_destroy (GtkWidget* widget, gpointer data)
   gtk_main_quit ();
 }
 
+static void
+main_window_realized (GtkWidget* widget, gpointer unused)
+{
+  GdkWindow *win = widget->window;
+
+  /* Some utilities search for our main window so that they can make
+     their dialogs transient for it.  They identify our window by
+     name.
+  */
+  XStoreName (GDK_WINDOW_XDISPLAY (win), GDK_WINDOW_XID (win),
+	      "osso-application-installer");
+}
+
 static void
 apt_status_callback (int cmd, apt_proto_decoder *dec, void *unused)
 {
   if (dec == NULL)
     return;
 
-  const char *label;
   int op = dec->decode_int ();
   int already = dec->decode_int ();
   int total = dec->decode_int ();
@@ -2418,6 +2561,18 @@ get_main_window ()
   return main_window;
 }
 
+GtkWidget *
+get_main_trail ()
+{
+  return main_trail;
+}
+
+GtkWidget *
+get_device_label ()
+{
+  return device_label;
+}
+
 void
 present_main_window ()
 {
@@ -2547,7 +2702,14 @@ enable_search (bool f)
 static void
 insensitive_press (GtkButton *button, gpointer data)
 {
-  irritate_user (_("ai_ib_not_available"));
+  char *text = (char *)data;
+  irritate_user (text);
+}
+
+static void
+insensitive_operation_press (GtkButton *button, gpointer data)
+{
+  irritate_user (insensitive_operation_press_label);
 }
 
 static osso_context_t *osso_ctxt;
@@ -2577,6 +2739,12 @@ set_current_help_topic (const char *topic)
   current_topic = topic;
 }
 
+static void
+call_refresh_package_cache (GtkWidget *button, gpointer data)
+{
+  refresh_package_cache (true);
+}
+
 int
 main (int argc, char **argv)
 {
@@ -2592,6 +2760,7 @@ main (int argc, char **argv)
   load_settings ();
 
   gtk_init (&argc, &argv);
+  setup_dbus();
 
   // XXX - We don't want a two-part title and this seems to be the
   //       only way to get rid of it.  Hopefully, setting an empty
@@ -2629,7 +2798,7 @@ main (int argc, char **argv)
 		    G_CALLBACK (do_current_operation),
 		    NULL);
   g_signal_connect (G_OBJECT (toolbar_operation_item), "insensitive_press",
-		    G_CALLBACK (insensitive_press), NULL);
+		    G_CALLBACK (insensitive_operation_press), NULL);
   gtk_toolbar_insert (GTK_TOOLBAR (toolbar),
 		      GTK_TOOL_ITEM (toolbar_operation_item),
 		      -1);
@@ -2647,17 +2816,12 @@ main (int argc, char **argv)
 		    G_CALLBACK (show_current_details),
 		    NULL);
   g_signal_connect (G_OBJECT (details_button), "insensitive_press",
-		    G_CALLBACK (insensitive_press), NULL);
+		    G_CALLBACK (insensitive_press),
+		    _("ai_ib_nothing_to_view"));
   gtk_toolbar_insert (GTK_TOOLBAR (toolbar),
 		      GTK_TOOL_ITEM (details_button),
 		      -1);
 
-#if 0
-  gtk_toolbar_insert (GTK_TOOLBAR (toolbar),
-		      GTK_TOOL_ITEM (gtk_separator_tool_item_new ()),
-		      -1);
-#endif
-
   image = gtk_image_new_from_icon_name ("qgn_toolb_gene_findbutton",
 					HILDON_ICON_SIZE_26);
   search_button = GTK_WIDGET (gtk_tool_button_new (image, NULL));
@@ -2667,11 +2831,27 @@ main (int argc, char **argv)
 		    G_CALLBACK (show_search_dialog),
 		    NULL);
   g_signal_connect (G_OBJECT (search_button), "insensitive_press",
-		    G_CALLBACK (insensitive_press), NULL);
+		    G_CALLBACK (insensitive_press),
+		    _("ai_ib_unable_search"));
   gtk_toolbar_insert (GTK_TOOLBAR (toolbar),
 		      GTK_TOOL_ITEM (search_button),
 		      -1);
 
+  if (red_pill_mode)
+    {
+      image = gtk_image_new_from_icon_name ("qgn_toolb_gene_refresh",
+					    HILDON_ICON_SIZE_26);
+      GtkWidget *button = GTK_WIDGET (gtk_tool_button_new (image, NULL));
+      gtk_tool_item_set_expand (GTK_TOOL_ITEM (button), TRUE);
+      gtk_tool_item_set_homogeneous (GTK_TOOL_ITEM (button), TRUE);
+      g_signal_connect (button, "clicked",
+			G_CALLBACK (call_refresh_package_cache),
+			NULL);
+      gtk_toolbar_insert (GTK_TOOLBAR (toolbar),
+			  GTK_TOOL_ITEM (button),
+			  -1);
+    }
+
   hildon_window_add_toolbar (HILDON_WINDOW (window),
 			     GTK_TOOLBAR (toolbar));
 
@@ -2686,6 +2866,9 @@ main (int argc, char **argv)
   g_signal_connect (G_OBJECT (window), "destroy",
 		    G_CALLBACK (window_destroy), NULL);
 
+  g_signal_connect (G_OBJECT (window), "realize",
+		    G_CALLBACK (main_window_realized), NULL);
+
   main_menu = GTK_MENU (gtk_menu_new ());
   create_menu (main_menu);
   hildon_window_set_menu (HILDON_WINDOW (window), GTK_MENU (main_menu));
