@@ -4,14 +4,13 @@
  * Parts of this file are derived from apt.  Apt is copyright 1997,
  * 1998, 1999 Jason Gunthorpe and others.
  *
- * Copyright (C) 2005 Nokia Corporation.
+ * Copyright (C) 2005, 2006 Nokia Corporation.  All Right reserved.
  *
  * Contact: Marius Vollmer <marius.vollmer@nokia.com>
  *
  * This program is free software; you can redistribute it and/or
- * modify it under the terms of the GNU General Public License as
- * published by the Free Software Foundation; either version 2 of the
- * License, or (at your option) any later version.
+ * modify it under the terms of the GNU General Public License version
+ * 2 as published by the Free Software Foundation.
  *
  * This program is distributed in the hope that it will be useful, but
  * WITHOUT ANY WARRANTY; without even the implied warranty of
@@ -40,7 +39,6 @@
 #include "util.h"
 
 #define _(x)       gettext (x)
-#define _UI2_(x,y) gettext_alt (x,y)
 
 struct repo_closure;
 
@@ -70,6 +68,7 @@ struct repo_closure {
   GtkTreeView *tree;
   GtkListStore *store;
 
+  GtkWindow *dialog;
   GtkWidget *edit_button;
   GtkWidget *delete_button;
 
@@ -195,16 +194,34 @@ repo_closure::find_repo (const char *deb)
 struct repo_edit_closure {
   bool isnew;
   bool readonly;
+  bool had_name;
   repo_line *line;
   GtkWidget *name_entry;
   GtkWidget *uri_entry;
   GtkWidget *dist_entry;
   GtkWidget *components_entry;
+
+  // Only one of these is non-NULL, depending on the UI version we
+  // implement.
+  //
   GtkWidget *enabled_button;
+  GtkWidget *disabled_button;
 };
 
 static void ask_the_pill_question ();
 
+/* Determine whether two STR1 equals STR2 when ignoring leading and
+   trailing whitespace in STR1.
+*/
+static bool
+stripped_equal (const char *str1, const char *str2)
+{
+  size_t len = strlen (str2);
+
+  str1 = skip_whitespace (str1);
+  return !strncmp (str1, str2, len) && all_white_space (str1 + len);
+}
+
 static void
 repo_edit_response (GtkDialog *dialog, gint response, gpointer clos)
 {
@@ -221,27 +238,48 @@ repo_edit_response (GtkDialog *dialog, gint response, gpointer clos)
       const char *dist = gtk_entry_get_text (GTK_ENTRY (c->dist_entry));
       const char *comps = gtk_entry_get_text (GTK_ENTRY (c->components_entry));
 
-      if (all_white_space (uri))
+      repo_line *r = c->line;
+      if (name)
+	{
+	  if (all_white_space (name))
+	    {
+	      if (c->had_name)
+		{
+		  irritate_user (_("ai_ib_enter_name"));
+		  gtk_widget_grab_focus (c->name_entry);
+		  return;
+		}
+	      name = NULL;
+	    }
+	  
+	  free (r->name);
+	  r->name = g_strdup (name);
+	}
+
+      if (all_white_space (uri) || stripped_equal (uri, "http://"))
 	{
 	  irritate_user (_("ai_ib_enter_web_address"));
+	  gtk_widget_grab_focus (c->uri_entry);
 	  return;
 	}
 
       if (all_white_space (dist))
 	{
 	  irritate_user (_("ai_ib_enter_distribution"));
+	  gtk_widget_grab_focus (c->dist_entry);
 	  return;
 	}
 
-      repo_line *r = c->line;
-      if (name && !all_white_space (name))
-	{
-	  free (r->name);
-	  r->name = g_strdup (name);
-	}
       free (r->line);
-      r->enabled =
-	gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (c->enabled_button));
+
+      if (c->enabled_button)
+	r->enabled =
+	  gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (c->enabled_button));
+      else
+	r->enabled =
+	  !gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON 
+					 (c->disabled_button));
+	
       r->line = g_strdup_printf ("%s %s %s %s",
 				 r->enabled? "deb" : "#deb", uri, dist, comps);
       r->deb_line = r->line + (r->enabled? 4 : 5);
@@ -265,26 +303,46 @@ static GtkWidget *
 add_entry (GtkWidget *box, GtkSizeGroup *group,
 	   const char *label,
 	   const char *text, const char *end,
-	   bool autocap, bool readonly)
+	   bool autocap, bool readonly, bool mandatory)
 {
   GtkWidget *caption, *entry;
   gint pos = 0;
 
-  entry = gtk_entry_new ();
-  g_object_set (entry, "autocap", autocap, NULL);
-  if (text)
-    gtk_editable_insert_text (GTK_EDITABLE (entry), text, end-text, &pos);
-  gtk_editable_set_editable (GTK_EDITABLE (entry), !readonly);
+  if (readonly)
+    {
+      GtkTextBuffer *buffer;
+
+      entry = gtk_text_view_new ();
+           
+      buffer = gtk_text_view_get_buffer (GTK_TEXT_VIEW (entry));
+      gtk_text_view_set_editable (GTK_TEXT_VIEW (entry), false);
+      
+      if (text)
+	gtk_text_buffer_set_text (buffer, text, end-text);
+
+      gtk_text_view_set_cursor_visible (GTK_TEXT_VIEW (entry), FALSE);
+      g_object_set (entry, "can-focus", FALSE, NULL);
+    }
+  else
+    {
+      entry = gtk_entry_new ();
+      g_object_set (entry, "autocap", autocap, NULL);
+      if (text)
+	gtk_editable_insert_text (GTK_EDITABLE (entry), text, end-text, &pos);
+    }
 
   caption = hildon_caption_new (group, label, entry,
-				NULL, HILDON_CAPTION_OPTIONAL);
+				NULL, (mandatory
+				       ? HILDON_CAPTION_MANDATORY
+				       : HILDON_CAPTION_OPTIONAL));
   gtk_box_pack_start_defaults (GTK_BOX (box), caption);
-
+  
   return entry;
 }
 
 static void
-show_repo_edit_dialog (repo_line *r, bool isnew, bool readonly)
+show_repo_edit_dialog (GtkWindow *parent,
+		       repo_line *r, bool isnew, bool readonly)
 {
   GtkWidget *dialog, *vbox, *caption;
   GtkSizeGroup *group;
@@ -294,26 +352,33 @@ show_repo_edit_dialog (repo_line *r, bool isnew, bool readonly)
   c->isnew = isnew;
   c->readonly = readonly;
   c->line = r;
-  
+  c->had_name = isnew;
+
   const char *title;
   if (readonly)
-    title = _UI2_("ai_ti_catalogue_details",
-		  "Catalogue details");
+    title = _("ai_ti_catalogue_details");
   else if (isnew)
     title = _("ai_ti_new_repository");
   else
     title = _("ai_ti_edit_repository");
 
   if (readonly)
-    dialog = gtk_dialog_new_with_buttons (title,
-					  get_main_window (),
-					  GTK_DIALOG_MODAL,
-					  "Close",
-					  1,
-					  NULL);
+    {
+      GtkWidget *button;
+
+      dialog = gtk_dialog_new_with_buttons (title, parent,
+					    GTK_DIALOG_MODAL,
+					    NULL);
+
+      // We create the button separately in order to put the focus on
+      // the button.
+      button = gtk_dialog_add_button (GTK_DIALOG (dialog),
+				      _("ai_bd_repository_close"),
+				      GTK_RESPONSE_CANCEL);
+      gtk_widget_grab_focus (button);
+    }
   else
-    dialog = gtk_dialog_new_with_buttons (title,
-					  get_main_window (),
+    dialog = gtk_dialog_new_with_buttons (title, parent,
 					  GTK_DIALOG_MODAL,
 					  _("ai_bd_new_repository_ok"),
 					  GTK_RESPONSE_OK,
@@ -321,6 +386,8 @@ show_repo_edit_dialog (repo_line *r, bool isnew, bool readonly)
 					  GTK_RESPONSE_CANCEL,
 					  NULL);
 
+  gtk_dialog_set_has_separator (GTK_DIALOG (dialog), FALSE);
+
   // XXX - there is no help for the "edit" version of this dialog.
   //
   if (isnew)
@@ -334,11 +401,13 @@ show_repo_edit_dialog (repo_line *r, bool isnew, bool readonly)
   if (ui_version > 1)
     {
       if (r->name)
-	end = r->name + strlen (r->name);
+	{
+	  end = r->name + strlen (r->name);
+	  c->had_name = true;
+	}
       c->name_entry = add_entry (vbox, group,
-				 _UI2_("ai_fi_new_repository_name",
-				       "Cataloque Name"),
-				 r->name, end, true, readonly);
+				 _("ai_fi_new_repository_name"),
+				 r->name, end, true, readonly, true);
     }
   else
     c->name_entry = NULL;
@@ -347,32 +416,49 @@ show_repo_edit_dialog (repo_line *r, bool isnew, bool readonly)
   parse_quoted_word (&start, &end, false);
   c->uri_entry = add_entry (vbox, group,
 			    _("ai_fi_new_repository_web_address"),
-			    start, end, false, readonly);
+			    start, end, false, readonly, true);
 
   start = end;
   parse_quoted_word (&start, &end, false);
   c->dist_entry = add_entry (vbox, group,
 			     _("ai_fi_new_repository_distribution"),
-			     start, end, false, readonly);
+			     start, end, false, readonly, true);
 
   start = end;
   parse_quoted_word (&start, &end, false);
   end = start + strlen (start);
   c->components_entry = add_entry (vbox, group,
 				   _("ai_fi_new_repository_component"),
-				   start, end, false, readonly);
+				   start, end, false, readonly, false);
 
-  c->enabled_button = gtk_check_button_new ();
-  gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (c->enabled_button),
-			       r->enabled);
-  caption = hildon_caption_new (group,
-				_("ai_fi_new_repository_enabled"),
-				c->enabled_button,
-				NULL, HILDON_CAPTION_OPTIONAL);
-  gtk_box_pack_start_defaults (GTK_BOX (vbox), caption);
-  gtk_widget_set_sensitive (c->enabled_button, !readonly);
+  if (ui_version < 2)
+    {
+      c->disabled_button = NULL;
+      c->enabled_button = gtk_check_button_new ();
+      gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (c->enabled_button),
+				    r->enabled);
+      caption = hildon_caption_new (group,
+				    _("ai_fi_new_repository_enabled"),
+				    c->enabled_button,
+				    NULL, HILDON_CAPTION_OPTIONAL);
+      gtk_box_pack_start_defaults (GTK_BOX (vbox), caption);
+      gtk_widget_set_sensitive (c->enabled_button, !readonly);
+    }
+  else
+    {
+      c->enabled_button = NULL;
+      c->disabled_button = gtk_check_button_new ();
+      gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (c->disabled_button),
+				    !(r->enabled));
+      caption = hildon_caption_new (group,
+				    _("ai_fi_new_repository_disabled"),
+				    c->disabled_button,
+				    NULL, HILDON_CAPTION_OPTIONAL);
+      gtk_box_pack_start_defaults (GTK_BOX (vbox), caption);
+      gtk_widget_set_sensitive (c->disabled_button, !readonly);
+    }
 
-  gtk_widget_set_usize (dialog, 400, -1);
+  gtk_widget_set_usize (dialog, 650, -1);
 
   g_signal_connect (dialog, "response",
 		    G_CALLBACK (repo_edit_response), c);
@@ -382,12 +468,12 @@ show_repo_edit_dialog (repo_line *r, bool isnew, bool readonly)
 static void
 add_new_repo (repo_closure *c)
 {
-  repo_line *r = new repo_line (c, "deb http:// mistral user", false, NULL);
+  repo_line *r = new repo_line (c, "deb http:// bora user", false, NULL);
   r->next = NULL;
   *c->lastp = r;
   c->lastp = &r->next;
 
-  show_repo_edit_dialog (r, true, false);
+  show_repo_edit_dialog (c->dialog, r, true, false);
 }
 
 static void
@@ -484,7 +570,9 @@ repo_response (GtkDialog *dialog, gint response, gpointer clos)
       if (r == NULL)
 	return;
 
-      show_repo_edit_dialog (r, false, false);
+      if (r->essential)
+	irritate_user (_("ai_ib_unable_edit"));
+      show_repo_edit_dialog (c->dialog, r, false, r->essential);
 
       return;
     }
@@ -510,7 +598,7 @@ repo_response (GtkDialog *dialog, gint response, gpointer clos)
       if (c->dirty)
 	{
 	  apt_worker_set_sources_list (repo_encoder, c, repo_reply, NULL);
-	  refresh_package_cache ();
+	  refresh_package_cache (true);
 	}
       
       delete c;
@@ -571,6 +659,7 @@ repo_row_activated (GtkTreeView *treeview,
 		    GtkTreeViewColumn *column,
 		    gpointer data)
 {
+  repo_closure *c = (repo_closure *)data;
   GtkTreeModel *model = gtk_tree_view_get_model (treeview);
   GtkTreeIter iter;
 
@@ -583,8 +672,7 @@ repo_row_activated (GtkTreeView *treeview,
 
       if (r->essential)
 	irritate_user (_("ai_ib_unable_edit"));
-      else
-	show_repo_edit_dialog (r, false, false);
+      show_repo_edit_dialog (c->dialog, r, false, r->essential);
     }
 }
 
@@ -602,7 +690,7 @@ repo_selection_changed (GtkTreeSelection *selection, gpointer data)
       if (r == NULL)
 	return;
 
-      gtk_widget_set_sensitive (c->edit_button, !r->essential);
+      gtk_widget_set_sensitive (c->edit_button, TRUE);
       gtk_widget_set_sensitive (c->delete_button, !r->essential);
     }
   else
@@ -616,13 +704,18 @@ static void
 refresh_repo_list (repo_closure *c)
 {
   gtk_list_store_clear (c->store);
+
+  gint position = 0;
   for (repo_line *r = c->lines; r; r = r->next)
     {
       if (r->deb_line)
 	{
 	  GtkTreeIter iter;
-	  gtk_list_store_append (c->store, &iter);
-	  gtk_list_store_set (c->store, &iter, 0, r, -1);
+	  gtk_list_store_insert_with_values (c->store, &iter,
+					     position,
+					     0, r,
+					     -1);
+	  position += 1;
 	}
     }
 }
@@ -656,7 +749,7 @@ make_repo_list (repo_closure *c)
 					      NULL);
 
   g_signal_connect (c->tree, "row-activated", 
-		    G_CALLBACK (repo_row_activated), NULL);
+		    G_CALLBACK (repo_row_activated), c);
 
   g_signal_connect
     (G_OBJECT (gtk_tree_view_get_selection (GTK_TREE_VIEW (c->tree))),
@@ -668,7 +761,7 @@ make_repo_list (repo_closure *c)
   scroller = gtk_scrolled_window_new (NULL, NULL);
   gtk_scrolled_window_set_policy (GTK_SCROLLED_WINDOW (scroller),
 				  GTK_POLICY_AUTOMATIC,
-				  GTK_POLICY_ALWAYS);
+				  GTK_POLICY_AUTOMATIC);
   gtk_container_add (GTK_CONTAINER (scroller), GTK_WIDGET (c->tree));
 
   return scroller;
@@ -682,18 +775,24 @@ maybe_add_new_repo_cont (bool res, void *data)
 
   if (res)
     {
-      repo_line *r = ac->new_repo;
-      ac->new_repo = NULL;
+      repo_line *old_repo = c->find_repo (ac->new_repo->deb_line);
 
-      r->clos = c;
-      r->next = NULL;
-      *c->lastp = r;
-      c->lastp = &r->next;
+      if (old_repo)
+	ac->cont (true, ac->cont_data);
+      else
+	{
+	  repo_line *r = ac->new_repo;
+	  ac->new_repo = NULL;
 
-      apt_worker_set_sources_list (repo_encoder, c, repo_reply, NULL);
+	  r->clos = c;
+	  r->next = NULL;
+	  *c->lastp = r;
+	  c->lastp = &r->next;
+	  
+	  apt_worker_set_sources_list (repo_encoder, c, repo_reply, NULL);
       
-      // Whooo, tail call elimination...
-      refresh_package_cache_with_cont (ac->cont, ac->cont_data);
+	  ac->cont (true, ac->cont_data);
+	}
     }
   else
     ac->cont (false, ac->cont_data);
@@ -707,13 +806,21 @@ maybe_add_new_repo_details (void *data)
 {
   repo_add_closure *ac = (repo_add_closure *)data;
 
-  show_repo_edit_dialog (ac->new_repo, true, true);
+  show_repo_edit_dialog (get_main_window (),
+			 ac->new_repo, true, true);
 }
 
 static void
-insensitive_press (GtkButton *button, gpointer data)
+insensitive_delete_press (GtkButton *button, gpointer data)
 {
-  irritate_user ((const gchar *)data);
+  repo_closure *c = (repo_closure *)data;
+
+  GtkTreeModel *model;
+  GtkTreeIter iter;
+
+  if (gtk_tree_selection_get_selected (gtk_tree_view_get_selection (c->tree),
+				       &model, &iter))
+    irritate_user (_("ai_ni_unable_remove_repository"));
 }
 
 void
@@ -750,15 +857,16 @@ sources_list_reply (int cmd, apt_proto_decoder *dec, void *data)
   c->lastp = rp;
   free (next_name);
 
+  /* XXX - do something with 'success'.
+   */
+
   int success = dec->decode_int ();
 
   repo_add_closure *ac = (repo_add_closure *)data;
 
   if (ac)
     {
-      repo_line *old_repo = c->find_repo (ac->new_repo->deb_line);
-
-      if (old_repo)
+      if (ac->for_install && c->find_repo (ac->new_repo->deb_line))
 	{
 	  ac->cont (true, ac->cont_data);
 
@@ -777,17 +885,13 @@ sources_list_reply (int cmd, apt_proto_decoder *dec, void *data)
 	  gchar *str;
 
 	  if (ac->for_install)
-	    str = g_strdup_printf (_UI2_("ai_ia_add_catalogue_text",
-					 "New catalogue needs to be added.\n"
-					 "Catalogue name: %s"),
+	    str = g_strdup_printf (_("ai_ia_add_catalogue_text"),
 				   name);
 	  else
-	    str = g_strdup_printf (_UI2_("ai_ia_add_catalogue_text2",
-					 "Add new catalogue?\n%s"),
+	    str = g_strdup_printf (_("ai_ia_add_catalogue_text2"),
 				   name);
 	    
-	  ask_yes_no_with_arbitrary_details (_UI2_("ai_ti_add_cataloque",
-						   "Add catalogue"),
+	  ask_yes_no_with_arbitrary_details (_("ai_ti_add_catalogue"),
 					     str,
 					     maybe_add_new_repo_cont,
 					     maybe_add_new_repo_details,
@@ -799,10 +903,13 @@ sources_list_reply (int cmd, apt_proto_decoder *dec, void *data)
     {
       GtkWidget *dialog = gtk_dialog_new ();
 
+      c->dialog = GTK_WINDOW (dialog);
+
       gtk_window_set_title (GTK_WINDOW (dialog), _("ai_ti_repository"));
       gtk_window_set_transient_for (GTK_WINDOW (dialog), get_main_window ());
       gtk_window_set_modal (GTK_WINDOW (dialog), TRUE);
       
+      gtk_dialog_set_has_separator (GTK_DIALOG (dialog), FALSE);
       gtk_dialog_add_button (GTK_DIALOG (dialog), 
 			     _("ai_bd_repository_new"), REPO_RESPONSE_NEW);
       c->edit_button = 
@@ -815,18 +922,19 @@ sources_list_reply (int cmd, apt_proto_decoder *dec, void *data)
 			     _("ai_bd_repository_close"), GTK_RESPONSE_CLOSE);
       respond_on_escape (GTK_DIALOG (dialog), GTK_RESPONSE_CLOSE);
       
-      g_signal_connect (c->edit_button, "insensitive_press",
-			G_CALLBACK (insensitive_press),
-			_("ai_ib_unable_edit"));
       g_signal_connect (c->delete_button, "insensitive_press",
-			G_CALLBACK (insensitive_press),
-			_("ai_ni_unable_remove_repository"));
+			G_CALLBACK (insensitive_delete_press), c);
       
       set_dialog_help (dialog, AI_TOPIC ("repository"));
       
       gtk_box_pack_start_defaults (GTK_BOX (GTK_DIALOG (dialog)->vbox),
 				   make_repo_list (c));
       
+      gtk_widget_set_sensitive (c->edit_button, FALSE);
+      gtk_widget_set_sensitive (c->delete_button, FALSE);
+
+      gtk_widget_set_usize (dialog, 0, 250);
+
       g_signal_connect (dialog, "response",
 			G_CALLBACK (repo_response), c);
       gtk_widget_show_all (dialog);
